# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

from logging import config as logging_config
from . import config

def init_config():
    log_file = config.get("LOG_FILE", "/var/log/funsun.log")
    log_level = config.get("LOG_LEVEL", "INFO")

    if log_file == '-':
        main_handler = {
            'level': log_level,
            'class': 'logging.StreamHandler',
            'formatter': 'main',
        }
    else:
        main_handler = {
            'level': log_level,
            'class': 'logging.FileHandler',
            'formatter': 'main',
            'filename': log_file,
        }

    log_config={
        "version":1,
        "formatters": {
            "main": {
                "format":'%(asctime)s %(process)d %(levelname)s %(message)s',
            }
        },
        "handlers": {
            "main": main_handler,
        },
        "loggers": {
            "": {
                'handlers': ['main'],
                'level': log_level
            },
            "tornado.application": {
                'handlers': ['main'],
                'level': log_level
            },
            "tornado.general": {
                'handlers': ['main'],
                'level': log_level
            },
            "tornadis.pool": {
                'handlers': ['main'],
                'level': "WARNING"
            },
            "tornadis.client": {
                'handlers': ['main'],
                'level': "WARNING"
            },
            "tornadis.connection": {
                'handlers': ['main'],
                'level': "WARNING"
            },
        }
    }

    logging_config.dictConfig(log_config)


# -*- coding: utf-8 -*-
# 18/1/25
# create by: snower

import time
import logging
from collections import defaultdict
import msgpack
from tornado import gen
from ...plan import Plan
from ..store import Store
from ...utils import unicode_type
from ... import config

class MemStore(Store):
    def __init__(self, *args, **kwargs):
        super(MemStore, self).__init__(*args, **kwargs)

        self.plans = {}
        self.time_plans = defaultdict(dict)
        self.current_time = int(time.mktime(time.gmtime()))
        self.store_file = config.get("STORE_MEM_STORE_FILE", "/tmp/forsun.session")

        self.load()
        logging.info("use mem store %s", self.store_file)

    def load(self):
        try:
            with open(self.store_file, "rb") as fp:
                data = fp.read()
                data = msgpack.loads(data)
                self.plans = {}
                for plan in data.get(b"plans", []):
                    plan = Plan.loads(plan)
                    self.plans[plan.key] = plan

                self.time_plans = defaultdict(dict)
                for key, plans in data.get(b"time_plans", {}).items():
                    self.time_plans[key] = {}
                    for plan_key in (plans or []):
                        if not isinstance(plan_key, unicode_type):
                            plan_key = unicode_type(plan_key, "utf-8")

                        if plan_key in self.plans:
                            self.time_plans[key][plan_key] = self.plans[plan_key]

                self.current_time = data.get(b"current_time", self.current_time)
            logging.info("store mem load session %s", self.store_file)
        except Exception as e:
            logging.error("store mem load session error: %s %s", self.store_file, e)

    def unload(self):
        try:
            with open(self.store_file, "wb") as fp:
                data = msgpack.dumps({
                    b"plans": [plan.dumps() for key, plan in self.plans.items()],
                    b"time_plans": {ts: list(plans.keys()) for ts, plans in self.time_plans.items()},
                    b"current_time": self.current_time,
                })
                fp.write(data)
            logging.info("store mem save session %s", self.store_file)
        except Exception as e:
            logging.error("store mem save session error: %s %s", self.store_file, e)

    @gen.coroutine
    def uninit(self):
        self.unload()

    @gen.coroutine
    def set_current(self, current_time):
        self.current_time = current_time
        raise gen.Return(self.current_time)

    @gen.coroutine
    def get_current(self):
        raise gen.Return(self.current_time)

    @gen.coroutine
    def set_plan(self, plan):
        self.plans[plan.key] = plan
        raise gen.Return(plan)

    @gen.coroutine
    def remove_plan(self, key):
        if key in self.plans:
            raise gen.Return(self.plans.pop(key))
        raise gen.Return(None)

    @gen.coroutine
    def get_plan(self, key):
        if key in self.plans:
            raise gen.Return(self.plans[key])
        raise gen.Return(None)

    @gen.coroutine
    def add_time_plan(self, plan):
        self.time_plans[plan.next_time][plan.key] = plan
        raise gen.Return(plan)

    @gen.coroutine
    def get_time_plan(self, ts):
        raise gen.Return(self.time_plans[ts].keys())

    @gen.coroutine
    def remove_time_plan(self, plan):
        if plan.key in self.time_plans[plan.next_time]:
            raise gen.Return(self.time_plans[plan.next_time].pop(plan.key))
        raise gen.Return(None)

    @gen.coroutine
    def delete_time_plan(self, ts):
        if ts in self.time_plans:
            raise gen.Return(self.time_plans.pop(ts))
        raise gen.Return(None)

    @gen.coroutine
    def get_plan_keys(self, prefix = ""):
        keys = []
        for key in self.plans:
            if key.startswith(prefix):
                keys.append(key)
        raise gen.Return(keys)


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

from __future__ import absolute_import, division, print_function, with_statement

import os
import sys
import argparse
import multiprocessing
import atexit
from ..forsun import config
from ..utils import is_py3

parser = argparse.ArgumentParser(description='High-performance timing scheduling service')
parser.add_argument('--bind', dest='bind_host', default="0.0.0.0", help='bind host (default: 0.0.0.0)')
parser.add_argument('--port', dest='bind_port', default=6458, type=int, help='bind port (default: 6458)')
parser.add_argument('--demon', dest='demon', nargs='?', const=True, default=False, type=bool, help='run demon mode')
parser.add_argument('--log', dest='log_file', default='/var/log/forsun.log', type=str, help='log file')
parser.add_argument('--log-level', dest='log_level', default='INFO', type=str, help='log level (defaul: INFO)')
parser.add_argument('--driver', dest='driver', default='mem', type=str, help='store driver mem or redis (defaul: mem)')
parser.add_argument('--driver-mem-store-file', dest='store_mem_store_file', default='/tmp/forsun.session', type=str, help='store mem driver store file (defaul: /tmp/forsun.session)')
parser.add_argument('--driver-redis-host', dest='driver_redis_host', default='127.0.0.1', type=str, help='store reids driver host (defaul: 127.0.0.1)')
parser.add_argument('--driver-redis-port', dest='driver_redis_port', default=6379, type=int, help='store reids driver port (defaul: 6379)')
parser.add_argument('--driver-redis-db', dest='driver_redis_db', default=0, type=int, help='store reids driver db (defaul: 0)')
parser.add_argument('--driver-redis-prefix', dest='driver_redis_prefix', default='forsun', type=str, help='store reids driver key prefix (defaul: forsun)')
parser.add_argument('--driver-redis-server-id', dest='driver_redis_server_id', default=0, type=int, help='store reids driver server id (defaul: 0)')
parser.add_argument('--extension-path', dest='extension_path', default='', type=str, help='extension path')
parser.add_argument('--extension', dest='extensions', default=[], action="append", type=str, help='extension name')

def serve(demon = True):
    if demon:
        sys.stdin.close()
        sys.stdin = open(os.devnull)

        sys.stdout.close()
        sys.stdout = open(os.devnull)

        sys.stderr.close()
        sys.stderr = open(os.devnull)

    from ..forsun import Forsun
    forsun = Forsun()
    forsun.serve()

def main():
    args = parser.parse_args()

    config.set("LOG_FILE", args.log_file)
    config.set("LOG_LEVEL", args.log_level)

    config.set("BIND_ADDRESS", args.bind_host)
    config.set("PORT", args.bind_port)

    config.set("STORE_DRIVER", args.driver)
    config.set("STORE_REDIS_HOST", args.driver_redis_host)
    config.set("STORE_REDIS_PORT", args.driver_redis_port)
    config.set("STORE_REDIS_DB", args.driver_redis_db)
    config.set("STORE_REDIS_PREFIX", args.driver_redis_prefix)
    config.set("STORE_REDIS_SERVER_ID", args.driver_redis_server_id)

    config.set("EXTENSION_PATH", args.extension_path)
    config.set("EXTENSIONS", args.extensions)


    if args.demon:
        p = multiprocessing.Process(target = serve, name=" ".join(sys.argv))
        p.start()
        if is_py3:
            atexit._clear()
        else:
            atexit._exithandlers = []
    else:
        serve(False)

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
# 15/6/8
# create by: snower

import time
import signal
import logging
try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty

__time_out_callback = None
__exit_callback = None
__queues = Queue()
__is_stop = False
__current_time = int(time.mktime(time.gmtime()))

def exit_handler(signum, frame):
    __queues.put((__exit_callback, tuple()), False)

def handler(signum, frame):
    __queues.put((__time_out_callback, (int(time.mktime(time.gmtime())),)), False)

def start(callback, exit_callback):
    global __time_out_callback, __exit_callback, __is_stop
    __is_stop = False
    __time_out_callback = callback
    __exit_callback = exit_callback

def stop():
    global __time_out_callback, __exit_callback, __is_stop
    __time_out_callback = None
    __exit_callback = None
    __is_stop = True
    logging.info("timer stoping")

def loop():
    signal.signal(signal.SIGHUP, exit_handler)
    signal.signal(signal.SIGINT, exit_handler)
    signal.signal(signal.SIGTERM, exit_handler)
    signal.signal(signal.SIGALRM, handler)
    signal.setitimer(signal.ITIMER_REAL, 1, 1)
    logging.info("timer ready")
    while not __is_stop:
        try:
            callback, args = __queues.get(True, 1)
            if callback:
                callback(*args)
        except Empty:
            continue

def current():
    return __current_time


# -*- coding: utf-8 -*-
# 15/6/10
# create by: snower

import logging
import threading
from tornado.ioloop import IOLoop
from thrift.protocol.TBinaryProtocol import TBinaryProtocolAcceleratedFactory
from torthrift.transport import TIOStreamTransportFactory
from torthrift.server import TTornadoServer
from .processor.Forsun import Processor
from .handler import Handler
from .. import timer
from .. import config

class ThriftServer(object):
    def __init__(self, forsun):
        self.forsun = forsun
        self.server = None
        self.thread = None

    def serve(self):
        handler = Handler(self.forsun)
        processor = Processor(handler)
        tfactory = TIOStreamTransportFactory()
        protocol = TBinaryProtocolAcceleratedFactory()

        bind_address = config.get("BIND_ADDRESS", "127.0.0.1")
        port = config.get("PORT", 6458)
        self.server = TTornadoServer(processor, tfactory, protocol)
        self.server.bind(port, bind_address)
        self.server.start(1)
        ioloop = IOLoop.instance()
        logging.info("starting server by %s:%s", bind_address, port)
        ioloop.start()

    def start(self):
        def _():
            try:
                self.serve()
            except Exception as e:
                logging.error("server error: %s", e)
                timer.stop()

        self.thread = threading.Thread(target=_)
        self.thread.setDaemon(True)
        self.thread.start()

    def stop(self):
        IOLoop.current().add_callback(lambda :IOLoop.current().stop())
        logging.info("server stoping")


# -*- coding: utf-8 -*-
# 15/6/27
# create by: snower

import os
from .utils import string_type, number_type

__config = {}

DEFAULT_CONFIG = {
    "LOG_FILE": "/var/log/funsun.log",
    "LOG_LEVEL": "ERROR",
    "LOG_FORMAT": "",

    "BIND_ADDRESS": "0.0.0.0",
    "PORT": 6458,

    "STORE_DRIVER": "mem",

    "STORE_MEM_STORE_FILE": "/tmp/forsun.session",

    "STORE_REDIS_HOST": "127.0.0.1",
    "STORE_REDIS_PORT": 6379,
    "STORE_REDIS_DB": 0,
    "STORE_REDIS_PREFIX": "forsun",
    "STORE_REDIS_SERVER_ID": 0,
    "STORE_REDIS_MAX_CONNECTIONS": 8,
    "STORE_REDIS_CLIENT_TIMEOUT": 7200,
    "STORE_REDIS_BULK_SIZE": 5,

    "ACTION_SHELL_CWD": "/tmp",
    "ACTION_HTTP_MAX_CLIENTS": 64,
    "ACTION_HTTP_CONNECT_TIMEOUT": 5,
    "ACTION_HTTP_REQUEST_TIMEOUT": 120,
    "ACTION_REDIS_MAX_CONNECTIONS": 8,
    "ACTION_REDIS_CLIENT_TIMEOUT": 7200,
    "ACTION_REDIS_BULK_SIZE": 5,
    "ACTION_THRIFT_MAX_CONNECTIONS": 64,
    "ACTION_MYSQL_USER": "root",
    "ACTION_MYSQL_PASSWD": "",
    "ACTION_MYSQL_MAX_CONNECTIONS": 8,
    "ACTION_MYSQL_WAIT_CONNECTION_TIMEOUT": 7200,
    "ACTION_MYSQL_IDLE_SECONDS": 120,

    "EXTENSION_PATH": "",
    "EXTENSIONS": [],
}

def get(name, default=None):
    return __config.get(name, default)

def set(name, value):
    old_value = __config[name]
    __config[name] = value
    return old_value

def update(config):
    __config.update(config)
    return __config

update(DEFAULT_CONFIG)
for key, value in DEFAULT_CONFIG.items():
    env_value = os.environ.get(key)
    if env_value is not None:
        try:
            if isinstance(value, number_type):
                set(key, int(env_value))
            elif isinstance(value, float):
                set(key, float(env_value))
            elif isinstance(value, string_type):
                set(key, str(env_value))
        except:pass


# -*- coding: utf-8 -*-
# 15/6/10
# create by: snower

import os
import sys
import logging
import traceback
from tornado.ioloop import IOLoop
from tornado import gen
from .servers import ThriftServer
from . import store
from . import action
from . import timer
from . import log
from .extension import ExtensionManager
from . import config
from . import error

class Forsun(object):
    def __init__(self):
        log.init_config()
        self.ioloop = IOLoop.current()

        store.init_stores()
        action.init_drivers()
        self.init_extensions()

        self.server = ThriftServer(self)
        self.store = store.get_store()
        self.current_time = None

    @gen.coroutine
    def init(self):
        ExtensionManager.init()
        yield self.store.init()

    @gen.coroutine
    def uninit(self):
        yield self.store.uninit()
        ExtensionManager.uninit()

    def init_extensions(self):
        extension_path = config.get("EXTENSION_PATH", [])
        if extension_path:
            sys.path.append(extension_path)
            logging.info("register extension path %s", extension_path)

        extensions = config.get("EXTENSIONS", [])
        for extension in extensions:
            try:
                extension_module, extension_class = "".join(extension.split(".")[:-1]), extension.split(".")[-1]
                extension_module = __import__(extension_module, globals(), [], [extension_class])
                extension_class = getattr(extension_module, extension_class)
                ExtensionManager.add_extension(extension_class)
                logging.info("load extension %s %s", extension, extension_class)
            except Exception as e:
                logging.error("load extension error: %s %s", extension, e)

        ExtensionManager.register()

    @gen.coroutine
    def execute_action(self, ts, plan):
        try:
            yield action.execute(ts, plan)
        except Exception as e:
            logging.error("plan %s action execute error: %s\n", plan.key, e, traceback.format_exc())

    @gen.coroutine
    def check_plan(self, ts, plan):
        plan.last_timeout = ts
        plan.current_count += 1
        plan.next_time = plan.get_next_time()
        if plan.next_time:
            yield self.store.add_time_plan(plan)
            yield self.store.set_plan(plan)
        else:
            yield self.store.remove_plan(plan.key)
            logging.debug("plan finish %s", plan.key)

    @gen.coroutine
    def handler_plan(self, ts, key):
        try:
            plan = yield self.store.get_plan(key)
            if plan:
                yield self.check_plan(ts, plan)
                self.ioloop.add_callback(self.execute_action, ts, plan)
        except Exception as e:
            logging.error("handler plan error: %s %s %s", ts, key, e)

    @gen.coroutine
    def handler(self, ts):
        try:
            plans = yield self.store.get_time_plan(ts)
            for key in plans:
                self.ioloop.add_callback(self.handler_plan, ts, key)
            yield self.store.delete_time_plan(ts)
        except Exception as e:
            logging.error("handler ts error: %s %s", ts, e)

    @gen.coroutine
    def check(self, ts):
        try:
            if self.current_time is None:
                self.current_time = yield self.store.get_current()
                logging.info("start by last time %s current time %s", self.current_time, timer.current())
                while self.current_time > 0 and self.current_time < ts:
                    yield self.handler(self.current_time)
                    self.current_time += 1
            self.current_time = ts
            yield self.store.set_current(self.current_time)
            yield self.handler(ts)
        except Exception as e:
            logging.error("check error: %s %s", ts, e)

    def time_out(self, ts):
        self.ioloop.add_callback(self.check, ts)

    @gen.coroutine
    def create_plan(self, plan):
        oplan = yield self.store.get_plan(plan.key)
        if oplan:
            yield self.store.remove_time_plan(oplan)
            yield self.store.remove_plan(oplan.key)
        if not plan.next_time:
            raise error.WillNeverArriveTimeError()

        try:
            action.get_driver(plan.action)
        except action.UnknownActionError:
            raise error.UnknownActionError()

        yield self.store.add_time_plan(plan)
        res = yield self.store.set_plan(plan)
        if not res:
            raise error.StorePlanError

    @gen.coroutine
    def remove_plan(self, key):
        oplan = yield self.store.get_plan(key)
        if not oplan:
            raise error.NotFoundPlanError()

        yield self.store.remove_time_plan(oplan)
        yield self.store.remove_plan(oplan.key)
        raise gen.Return(oplan)

    @gen.coroutine
    def get_pan(self, key):
        plan = yield self.store.get_plan(key)
        if not plan:
            raise error.NotFoundPlanError()
        raise gen.Return(plan)

    @gen.coroutine
    def get_keys(self, prefix=""):
        keys = yield self.store.get_plan_keys(prefix)
        raise gen.Return(keys)

    @gen.coroutine
    def get_current_time(self):
        raise gen.Return(timer.current())

    @gen.coroutine
    def get_time_plans(self, ts):
        keys = yield self.store.get_time_plan(ts)
        plans = []
        for key in keys:
            plan = yield self.store.get_plan(key)
            if plan:
                plans.append(plan)
        raise gen.Return(plans)

    def serve(self):
        try:
            self.ioloop.add_callback(self.init)
            self.ioloop.add_callback(logging.info, "forsun ready %s", os.getpid())
            self.server.start()
            timer.start(self.time_out, self.exit)
            timer.loop()
        except KeyboardInterrupt:
            self.exit()

    def exit(self):
        @gen.coroutine
        def on_exit():
            yield self.uninit()
            self.server.stop()
            timer.stop()
            logging.info("stoped current time %s", timer.current())
        self.ioloop.add_callback(on_exit)
        logging.info("stoping")